/*
 * Mutex for two contenders
 */

module mutex2bit(CLOCK, in0, in1, out0, out1);
   input  CLOCK;
   input  in0, in1; // request lines
   output out0, out1;
   reg    out0, out1; // granted lines
   reg    last_state; // hi if serving a request, lo if not

   initial begin
      last_state = 1'b0;
   end

   always @(posedge CLOCK) begin
      if (last_state == 1'b0) begin
         last_state <= 1'b1;
         if (in0 == 1'b1) begin
            out0 <= 1'b1;
            out1 <= 1'b0;
         end
         else if (in1 == 1'b1) begin
            out0 <= 1'b0;
            out1 <= 1'b1;
         end
      end // if (last_state == 1'b0)
      else begin
         if (// if serving out0 and in0 is lo
             (out0 == 1'b1 & out1 == 1'b0 & in0 == 1'b0) |
             // if serving out1 and in1 is lo
             (out0 == 1'b0 & out1 == 1'b1 & in1 == 1'b0)) begin
            last_state <= 1'b0;
            out0 <= 1'b0;
            out1 <= 1'b0;
         end
      end
   end
endmodule // mutex2bit

                                             
                 
